@@ -6,7 +6,7 @@ subst - Greple module for text search and substitution
 
 =head1 VERSION
 
-Version 2.28
+Version 2.29
 
 =head1 SYNOPSIS
 
@@ -22,6 +22,7 @@ greple -Msubst --dict I<dictionary> [ options ]
   --with-stat
   --stat-style=[default,dict]
   --stat-item={match,expect,number,ok,ng}=[0,1]
+  --subst
   --diff
   --diffcmd command
   --create
@@ -64,6 +65,9 @@ line.
 
     greple --dictdata $'colou?r color\ncent(er|re) center\n'
 
+Dictionary entry starting with a sharp sign (C<#>) is comment and
+ignored.
+
 =head2 Overlapped pattern
 
 When the matched string is same or shorter than previously matched
@@ -142,6 +146,7 @@ are:
     number=1
     ng=1
     ok=1
+    dict=0
 
 If you don't need to see pattern field, use like this:
 
@@ -320,7 +325,7 @@ it under the same terms as Perl itself.
 use v5.14;
 package App::Greple::subst;
 
-our $VERSION = '2.28';
+our $VERSION = '2.29';
 
 use warnings;
 use utf8;
@@ -373,7 +378,10 @@ our $opt_warn_overlap = 1;
 our $opt_warn_include = 0;
 our $opt_stat_style = "default";
 our @opt_stat_item;
-our %opt_stat_item = map { $_ => 1 } qw(match expect number ng ok);
+our %opt_stat_item = (
+    map( { $_ => 1 } qw(match expect number ng ok) ),
+    map( { $_ => 0 } qw(dict) ),
+    );
 our $opt_show_comment = 0;
 our $opt_show_numbers = 1;
 my %stat;
@@ -382,7 +390,7 @@ my $current_file;
 my $contents;
 my @subst_diffcmd;
 my $ignorechar_re;
-my $dict = App::Greple::subst::Dict->new;
+my @dicts;
 
 sub debug {
     $debug = 1;
@@ -406,22 +414,27 @@ sub subst_initialize {
 	@subst_diffcmd = ("diff", "-U$opt_U");
     }
 
+    my $config = { linefold  => $opt_linefold,
+		   dictname  => $opt_dictname,
+		   printdict => $opt_printdict };
     for my $data (@opt_dictdata) {
-	if (utf8::is_utf8 $data) {
-	    $data = encode 'utf8', $data;
-	}
-	open my $fh, "<", \$data;
-	read_dict_fh($fh);
+	push @dicts, App::Greple::subst::Dict->new(
+	    DATA => $data,
+	    CONFIG => $config,
+	    );
     }
     for my $file (@opt_dictfile) {
 	if (-d $file) {
 	    warn "$file is directory\n";
 	    next;
 	}
-	read_dict($file);
+	push @dicts, App::Greple::subst::Dict->new(
+	    FILE => $file,
+	    CONFIG => $config,
+	    );
     }
 
-    if ($dict->words == 0) {
+    if (@dicts == 0) {
 	warn "Module -Msubst requires dictionary data.\n";
 	main::usage();
 	die;
@@ -458,7 +471,7 @@ sub subst_begin {
 
 use Text::VisualWidth::PP;
 use Text::VisualPrintf qw(vprintf vsprintf);
-use List::Util qw(max any);
+use List::Util qw(max any sum);
 
 sub vwidth {
     if (not defined $_[0] or length $_[0] == 0) {
@@ -471,30 +484,36 @@ my @match_list;
 
 sub subst_show_stat {
     my %arg = @_;
-    my @fromto = $dict->words;
     my($from_max, $to_max) = (0, 0);
-    my @show;
-    for my $i (0 .. $#fromto) {
-	my $p = $fromto[$i] // next;
-	if ($p->is_comment) {
-	    push @show, [ $i, $p, {} ];
-	    next;
+    my $i = -1;
+    my @show_list;
+    for my $dict (@dicts) {
+	my @fromto = $dict->words;
+	my @show;
+	for my $p (@fromto) {
+	    $i++;
+	    $p // die;
+	    if ($p->is_comment) {
+		push @show, [ $i, $p, {} ] if $opt_show_comment;
+		next;
+	    }
+	    my($from_re, $to) = ($p->string, $p->correct // '');
+	    my $hash = $match_list[$i] // {};
+	    my @keys = keys %{$hash};
+	    my @ng = grep { $_ ne $to } @keys;
+	    my @ok = grep { $_ eq $to } @keys;
+	    if    ($opt_check eq 'none'    ) { next if @keys != 0 }
+	    elsif ($opt_check eq 'any'     ) { next if @keys == 0 }
+	    elsif ($opt_check eq 'ok'      ) { next if @ok   == 0 }
+	    elsif ($opt_check eq 'ng'      ) { next if @ng   == 0 }
+	    elsif ($opt_check eq 'outstand') { next if @ng   == 0 }
+	    elsif ($opt_check eq 'all')      { }
+	    else { die }
+	    $from_max = max $from_max, vwidth $from_re;
+	    $to_max   = max $to_max  , vwidth $to;
+	    push @show, [ $i, $p, $hash ];
 	}
-	my($from_re, $to) = ($p->string, $p->correct // '');
-	my $hash = $match_list[$i] // {};
-	my @keys = keys %{$hash};
-	my @ng = grep { $_ ne $to } @keys;
-	my @ok = grep { $_ eq $to } @keys;
-	if    ($opt_check eq 'none'    ) { next if @keys != 0 }
-	elsif ($opt_check eq 'any'     ) { next if @keys == 0 }
-	elsif ($opt_check eq 'ok'      ) { next if @ok   == 0 }
-	elsif ($opt_check eq 'ng'      ) { next if @ng   == 0 }
-	elsif ($opt_check eq 'outstand') { next if @ng   == 0 }
-	elsif ($opt_check eq 'all')      { }
-	else { die }
-	$from_max = max $from_max, vwidth $from_re;
-	$to_max   = max $to_max  , vwidth $to;
-	push @show, [ $i, $p, $hash ];
+	push @show_list, [ $dict => \@show ];
     }
     if ($opt_show_numbers) {
 	no warnings 'uninitialized';
@@ -502,63 +521,43 @@ sub subst_show_stat {
 	    $stat{hit}, $stat{total},
 	    $stat{ng}, $stat{ok}, $stat{ng} + $stat{ok};
     }
-    for my $show (@show) {
-	my($i, $p, $hash) = @$show;
-	if ($p->is_comment) {
-	    say $p->comment if $opt_show_comment;
-	    next;
-	}
-	my($from_re, $to) = ($p->string, $p->correct // '');
-	my @keys = keys %{$hash};
-	if ($opt_stat_style eq 'dict') {
-	    vprintf("%-${from_max}s // %s", $from_re // '', $to // '');
-	} else {
-	    my @ng = sort { $hash->{$b} <=> $hash->{$a} } grep { $_ ne $to } @keys
-		if $opt_stat_item{ng};
-	    my @ok = grep { $_ eq $to } @keys
-		if $opt_stat_item{ok};
-	    vprintf("%${from_max}s => ", $from_re // '') if $opt_stat_item{match};
-	    vprintf("%-${to_max}s",      $to // '')      if $opt_stat_item{expect};
-	    vprintf(" %4d:",             $i + 1)         if $opt_stat_item{number};
-	    for my $key (@ng, @ok) {
-		my $index = $key eq $to ? $i * 2 + 1 : $i * 2;
-		printf(" %s(%s)",
-		       main::index_color($index, $key),
-		       colorize($key eq $to ? 'DB' : 'DR', $hash->{$key})
-		    );
+    for my $show_list (@show_list) {
+	my($dict, $show) = @{$show_list};
+	next if @$show == 0;
+	my $dict_format = "[ %s ]\n";
+	printf $dict_format, $dict->NAME if $opt_stat_item{dict};
+	for my $item (@$show) {
+	    my($i, $p, $hash) = @$item;
+	    if ($p->is_comment) {
+		say $p->comment if $opt_show_comment;
+		next;
 	    }
+	    my($from_re, $to) = ($p->string, $p->correct // '');
+	    my @keys = keys %{$hash};
+	    if ($opt_stat_style eq 'dict') {
+		vprintf("%-${from_max}s // %s", $from_re // '', $to // '');
+	    } else {
+		my @ng = sort { $hash->{$b} <=> $hash->{$a} } grep { $_ ne $to } @keys
+		    if $opt_stat_item{ng};
+		my @ok = grep { $_ eq $to } @keys
+		    if $opt_stat_item{ok};
+		vprintf("%${from_max}s => ", $from_re // '') if $opt_stat_item{match};
+		vprintf("%-${to_max}s",      $to // '')      if $opt_stat_item{expect};
+		vprintf(" %4d:",             $i + 1)         if $opt_stat_item{number};
+		for my $key (@ng, @ok) {
+		    my $index = $key eq $to ? $i * 2 + 1 : $i * 2;
+		    printf(" %s(%s)",
+			   main::index_color($index, $key),
+			   colorize($key eq $to ? 'DB' : 'DR', $hash->{$key})
+			);
+		}
+	    }
+	    print "\n";
 	}
-	print "\n";
     }
     $_ = "";
 }
 
-sub read_dict {
-    my $file = shift;
-    say $file if $opt_dictname;
-    open my $fh, "<", $file or die "$file: $!\n";
-    read_dict_fh($fh);
-}
-
-sub read_dict_fh {
-    my $fh = shift;
-    local $_;
-    my $flag = FLAG_REGEX;
-    $flag |= FLAG_COOK if $opt_linefold;
-    while (<$fh>) {
-	chomp;
-	say if $opt_printdict;
-	if (not /^\s*[^#]/) {
-	    $dict->add_comment($_);
-	    next;
-	}
-	my @param = grep { not m{^//+$} } split ' ';
-	splice @param, 0, -2; # leave last one or two
-	my($pattern, $correct) = @param;
-	$dict->add($pattern, $correct, flag => $flag);
-    }
-}
-
 use App::Greple::Regions qw(match_regions merge_regions filter_regions);
 use List::MoreUtils qw(pairwise);
 
@@ -570,76 +569,78 @@ sub subst_search {
     my @matched;
     my $index = -1;
     my @effective;
-    my $ng = any { $opt_check eq $_ } qw(ng any all none);
-    my $ok = any { $opt_check eq $_ } qw(ok any all none);
+    my $ng = {ng=>1, any=>1, all=>1, none=>1}->{$opt_check} ;
+    my $ok = {ok=>1, any=>1, all=>1, none=>1}->{$opt_check} ;
     my $outstand = $opt_check eq 'outstand';
-    for my $p ($dict->words) {
-	$index++;
-	$p // next;
-	next if $p->is_comment;
-	my($from_re, $to) = ($p->string, $p->correct // '');
-	my @match = match_regions pattern => $p->regex;
+    for my $dict (@dicts) {
+	for my $p ($dict->words) {
+	    $index++;
+	    $p // next;
+	    next if $p->is_comment;
+	    my($from_re, $to) = ($p->string, $p->correct // '');
+	    my @match = match_regions pattern => $p->regex;
 
-	##
-	## Remove all overlapped matches.
-	##
-	my($in, $over, $out, $im, $om) = filter_regions \@match, \@matched;
-	@match = @$out;
-	for my $warn (
-	    [ "Include", $in,   $im, $opt_warn_include ],
-	    [ "Overlap", $over, $om, $opt_warn_overlap ],
-	    ) {
-	    my($kind, $list, $match, $show) = @$warn;
-	    $show and @$list or next;
-	    pairwise {
-		warn sprintf("%s \"%s\" with \"%s\" by #%d /%s/ in %s at %d\n",
-			     $kind,
-			     substr($_, $a->[0], $a->[1] - $a->[0]),
-			     substr($_, $b->[0], $b->[1] - $b->[0]),
-			     $index + 1, $p->string,
-			     $current_file,
-			     $a->[0],
-		    );
-	    } @$list, @$match;
-	}
-
-	$stat{total}++;
-	$stat{hit}++ if @match;
-	next if @match == 0 and $opt_check ne 'all';
-
-	my $hash = $match_list[$index] //= {};
-	my $callback = sub {
-	    my($ms, $me, $i, $matched) = @_;
-	    $stat{$i % 2 ? 'ok' : 'ng'}++;
-	    my $s = $matched =~ s/$ignorechar_re//gr;
-	    $hash->{$s}++;
-	    my $format = @opt_format[ $i % @opt_format ];
-	    sprintf($format,
-		    ($opt_subst && $to ne '' && $s ne $to) ?
-		    $to : $matched);
-	};
-	my(@ok, @ng);
-	for (@match) {
-	    my $matched = substr $text, $_->[0], $_->[1] - $_->[0];
-	    if ($matched =~ s/$ignorechar_re//gr ne $to) {
-		$_->[2] = $index * 2;
-		push @ng, $_;
-	    } else {
-		$_->[2] = $index * 2 + 1;
-		push @ok, $_;
+	    ##
+	    ## Remove all overlapped matches.
+	    ##
+	    my($in, $over, $out, $im, $om) = filter_regions \@match, \@matched;
+	    @match = @$out;
+	    for my $warn (
+		[ "Include", $in,   $im, $opt_warn_include ],
+		[ "Overlap", $over, $om, $opt_warn_overlap ],
+		) {
+		my($kind, $list, $match, $show) = @$warn;
+		$show and @$list or next;
+		pairwise {
+		    warn sprintf("%s \"%s\" with \"%s\" by #%d /%s/ in %s at %d\n",
+				 $kind,
+				 substr($_, $a->[0], $a->[1] - $a->[0]),
+				 substr($_, $b->[0], $b->[1] - $b->[0]),
+				 $index + 1, $p->string,
+				 $current_file,
+				 $a->[0],
+			);
+		} @$list, @$match;
 	    }
-	    $_->[3] = $callback;
-	}
-	$effective[ $index * 2     ] = 1 if $ng || ( @ng && $outstand );
-	$effective[ $index * 2 + 1 ] = 1 if $ok || ( @ng && $outstand );
 
-	@matched = merge_regions { nojoin => 1 }, @matched, @match;
+	    $stat{total}++;
+	    $stat{hit}++ if @match;
+	    next if @match == 0 and $opt_check ne 'all';
+
+	    my $hash = $match_list[$index] //= {};
+	    my $callback = sub {
+		my($ms, $me, $i, $matched) = @_;
+		$stat{$i % 2 ? 'ok' : 'ng'}++;
+		my $s = $matched =~ s/$ignorechar_re//gr;
+		$hash->{$s}++;
+		my $format = @opt_format[ $i % @opt_format ];
+		sprintf($format,
+			($opt_subst && $to ne '' && $s ne $to) ?
+			$to : $matched);
+	    };
+	    my(@ok, @ng);
+	    for (@match) {
+		my $matched = substr $text, $_->[0], $_->[1] - $_->[0];
+		if ($matched =~ s/$ignorechar_re//gr ne $to) {
+		    $_->[2] = $index * 2;
+		    push @ng, $_;
+		} else {
+		    $_->[2] = $index * 2 + 1;
+		    push @ok, $_;
+		}
+		$_->[3] = $callback;
+	    }
+	    $effective[ $index * 2     ] = 1 if $ng || ( @ng && $outstand );
+	    $effective[ $index * 2 + 1 ] = 1 if $ok || ( @ng && $outstand );
+
+	    @matched = merge_regions { nojoin => 1 }, @matched, @match;
+	}
     }
     ##
     ## --select
     ##
     if (my $select = $opt_subst_select) {
-	my $max = $dict->words;
+	my $max = sum map { int $_->words } @dicts;
 	use Getopt::EX::Numbers;
 	my $numbers = Getopt::EX::Numbers->new(min => 1, max => $max);
 	my @select;
